class CaseData:
    """
    A :class:`.PeriodicInfo` test case.

    Attributes
    ----------
    periodic_info : :class:`.PeriodicInfo`
        The information to test.

    vector_1 : :class:`numpy.ndarray`
        Cell lattice vector of shape (3, ) in x direction in
        Angstrom.

    vector_2 : :class:`numpy.ndarray`
        Cell lattice vector of shape (3, ) in y direction in
        Angstrom.

    vector_3 : :class:`numpy.ndarray`
        Cell lattice vector of shape (3, ) in z direction in
        Angstrom.

    a : :class:`float`
        Length in a direction of cell matrix.

    b : :class:`float`
        Length in b direction of cell matrix.

    c : :class:`float`
        Length in c direction of cell matrix.

    alpha : :class:`float`
        Alpha angle of cell matrix.

    beta : :class:`float`
        Beta angle of cell matrix.

    gamma : :class:`float`
        Gamma angle of cell matrix.

    """

    def __init__(
        self,
        periodic_info,
        vector_1,
        vector_2,
        vector_3,
        a,
        b,
        c,
        alpha,
        beta,
        gamma,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        periodic_info : :class:`.PeriodicInfo`
            The information to test.

        vector_1 : :class:`numpy.ndarray`
            Cell lattice vector of shape (3, ) in x direction in
            Angstrom.

        vector_2 : :class:`numpy.ndarray`
            Cell lattice vector of shape (3, ) in y direction in
            Angstrom.

        vector_3 : :class:`numpy.ndarray`
            Cell lattice vector of shape (3, ) in z direction in
            Angstrom.

        a : :class:`float`
            Length in a direction of cell matrix.

        b : :class:`float`
            Length in b direction of cell matrix.

        c : :class:`float`
            Length in c direction of cell matrix.

        alpha : :class:`float`
            Alpha angle of cell matrix.

        beta : :class:`float`
            Beta angle of cell matrix.

        gamma : :class:`float`
            Gamma angle of cell matrix.

        """

        self.periodic_info = periodic_info
        self.vector_1 = vector_1
        self.vector_2 = vector_2
        self.vector_3 = vector_3
        self.a = a
        self.b = b
        self.c = c
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma



import numpy as np
import stk

from ...utilities import (
    get_displacement_vector,
    has_same_structure,
    is_clone,
)


def test_with_rotation_between_vectors(molecule, target, get_origin):
    """
    Test :meth:`.Molecule.with_rotation_between_vectors`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    target : :class:`numpy.ndarray`
        The target vector which defines the applied rotation.

    get_origin : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `origin` parameter for this test. Using a :class:`callable`
        allows the origin for the test to be set to molecule-specific
        values, such as the centroid of the molecule being tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Use to check that immutability is not violated.
    clone = molecule.clone()
    _test_with_rotation_between_vectors(molecule, target, get_origin)
    has_same_structure(molecule, clone)


def _test_with_rotation_between_vectors(molecule, target, get_origin):
    """
    Test :meth:`.Molecule.with_rotation_between_vectors`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    target : :class:`numpy.ndarray`
        The target vector which defines the applied rotation.

    get_origin : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `origin` parameter for this test. Using a :class:`callable`
        allows the origin for the test to be set to molecule-specific
        values, such as the centroid of the molecule being tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    start = get_displacement_vector(molecule, 0, 1)
    new = molecule.with_rotation_between_vectors(
        start=start,
        target=target,
        origin=get_origin(molecule),
    )
    is_clone(new, molecule)
    result = get_displacement_vector(new, 0, 1)
    assert np.allclose(
        a=stk.normalize_vector(result),
        b=stk.normalize_vector(target),
        atol=1e-12,
    )
    assert abs(np.linalg.norm(start) - np.linalg.norm(result)) < 1e-14



from __future__ import annotations

from functools import partial

import numpy as np
import pytest
import stk
from scipy.spatial.distance import euclidean

from ....case_data import CaseData


@pytest.fixture
def angled(
    position: np.ndarray,
    angled_aligner_edge: int,
    building_block_2: stk.BuildingBlock,
) -> CaseData:
    point1, point2 = points = (
        position + [0, 10, 0],
        position + [10, 0, 0],
    )

    def get_fg0_point(building_block: stk.BuildingBlock) -> float:
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_fg1_point(building_block: stk.BuildingBlock) -> float:
        return get_closest_point(
            points=points,
            point=get_fg_position(1, building_block),
        )

    vertex = stk.cage.AngledVertex(
        id=0,
        position=position,
        aligner_edge=angled_aligner_edge,
    )

    return CaseData(
        vertex=vertex,
        edges=tuple(get_angled_edges(vertex)),
        building_block=building_block_2,
        position=position,
        alignment_tests={
            get_fg0_point: (point1 if angled_aligner_edge == 0 else point2),
            get_fg1_point: (point2 if angled_aligner_edge == 0 else point1),
        },
        functional_group_edges=(
            {0: 0, 1: 1} if angled_aligner_edge == 0 else {0: 1, 1: 0}
        ),
    )


def get_closest_point(
    points: tuple[np.ndarray, ...],
    point: np.ndarray,
) -> float:
    return min(points, key=partial(euclidean, point))


def get_fg_position(
    id: id,
    building_block: stk.BuildingBlock,
) -> np.ndarray:
    functional_group = next(building_block.get_functional_groups(id))
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_angled_edges(vertex: stk.Vertex) -> stk.Edge:
    vertex2 = stk.Vertex(1, vertex.get_position() + [0, 10, 0])
    vertex3 = stk.Vertex(2, vertex.get_position() + [10, 0, 0])
    yield stk.Edge(0, vertex, vertex2)
    yield stk.Edge(1, vertex, vertex3)


@pytest.fixture(params=(0, 1))
def angled_aligner_edge(request) -> int:
    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCCNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_2(request) -> stk.BuildingBlock:
    return request.param()


@pytest.fixture(
    params=([1, 2, -20],),
)
def position(request) -> np.ndarray:
    """
    The `position` of a vertex.

    """

    return np.array(request.param, dtype=np.float64)



"""
Serial Evolutionary Algorithm
=============================

"""

from .implementation import Implementation


class Serial(Implementation):
    """
    A serial implementation of the default evolutionary algorithm.

    """

    def get_generations(self, num_generations):
        yield from self._get_generations(num_generations, map)



from stk._internal.molecule import Molecule

from .molecule import MoleculeKeyMaker
from .utilities import get_smiles


class Smiles(MoleculeKeyMaker):
    """
    Used to get the SMILES of molecules.

    Examples:

        *Adding SMILES to a Molecule's JSON*

        You want to use the isomeric, canonical SMILES from RDKit as
        part of a JSON representation of a molecule

        .. testcode:: adding-smiles-to-a-molecules-json

            import stk

            jsonizer = stk.MoleculeJsonizer(
                key_makers=(stk.Smiles(), ),
            )
            # Get the JSON representation, including an SMILES.
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        .. testcode:: adding-smiles-to-a-molecules-json
            :hide:

            assert json['molecule']['SMILES'] == 'NCCN'
            assert json['matrix']['SMILES'] == 'NCCN'
    """

    def __init__(self) -> None:
        return

    def get_key_name(self) -> str:
        return "SMILES"

    def get_key(self, molecule: Molecule) -> str:
        return get_smiles(molecule)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return "Smiles()"



import pytest
import stk

from ..case_data import CaseData


def _get_case_data_1() -> CaseData:
    topology_graph = stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=2,
    )

    return CaseData.new(
        fitness_normalizer=stk.NormalizerSequence(
            fitness_normalizers=(
                stk.Multiply(
                    coefficient=(1, 2, 4),
                    filter=lambda fitness_values, record: fitness_values[
                        record
                    ]
                    is not None,
                ),
                stk.Sum(
                    filter=lambda fitness_values, record: fitness_values[
                        record
                    ]
                    is not None,
                ),
            ),
        ),
        fitness_values={
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((4, 10, 1), 28),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((9, 20, 2), 57),
            stk.MoleculeRecord(
                topology_graph=topology_graph,
            ): ((16, 30, 4), 92),
            stk.MoleculeRecord(topology_graph): (None, None),
        },
    )


@pytest.fixture(
    scope="session",
    params=(_get_case_data_1,),
)
def sequence(request: pytest.FixtureRequest) -> CaseData:
    return request.param()



import logging
import typing
from collections.abc import Iterable, Iterator

from stk._internal.ea.crossover.molecule_crosser import MoleculeCrosser
from stk._internal.ea.fitness_calculators.fitness_calculator import (
    FitnessCalculator,
)
from stk._internal.ea.fitness_normalizers.fitness_normalizer import (
    FitnessNormalizer,
)
from stk._internal.ea.generation import Generation
from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.mutation.mutator import MoleculeMutator
from stk._internal.ea.selection.selectors.selector import Selector
from stk._internal.key_makers.inchi import Inchi
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from ..fitness_normalizers.null import NullFitnessNormalizer
from .implementations.parallel import Parallel
from .implementations.serial import Serial

logger = logging.getLogger(__name__)

T = typing.TypeVar("T", bound=MoleculeRecord)


class EvolutionaryAlgorithm(typing.Generic[T]):
    """
    An abstract base class for evolutionary algorithms.

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. This is purely for convenience, so that
        there is a default evolutionary algorithm implementation that
        users can use. However, feel free to override the default
        implementation when implementing subclasses.

        If you do want to use the default implementation, here is a
        summary of the roles of the different components:

        .. image:: https://i.imgur.com/hGXboaU.png

    Examples:

        *Subclass Implementation*

        The source code of this class can work as a good example. There
        is only one method that a subclass of
        :class:`.EvolutionaryAlgorithm` needs to implement,
        :meth:`.get_generations`, which yields :class:`.Generation`
        instances. These correspond to the generations of your
        evolutionary algorithm implementation.

        *Usage*

        There are a couple of tutorials on how to use the
        :class:`.EvolutionaryAlgorithm`, which can be found in the sidebar.

    """

    _implementation: Serial | Parallel

    def __init__(
        self,
        initial_population: Iterable[T],
        fitness_calculator: FitnessCalculator[T],
        mutator: MoleculeMutator[T],
        crosser: MoleculeCrosser[T],
        generation_selector: Selector[T],
        mutation_selector: Selector[T],
        crossover_selector: Selector[T],
        fitness_normalizer: FitnessNormalizer[T] = NullFitnessNormalizer(),
        key_maker: MoleculeKeyMaker = Inchi(),
        num_processes: int | None = None,
    ) -> None:
        """
        Parameters:

            initial_population (list[T]):
                The initial population the EA should use.

            fitness_calculator:
                Calculates fitness values.

            mutator:
                Carries out mutation operations.

            crosser:
                Carries out crossover operations.

            generation_selector:
                Selects the next generation.

            mutation_selector:
                Selects molecules for mutation.

            crossover_selector:
                Selects molecules for crossover.

            fitness_normalizer:
                Normalizes fitness values.

            key_maker:
                Used to detect duplicate molecules in the EA. If two
                molecules in a generation return the same key, one of them
                is removed.

            num_processes:
                The number of parallel processes the EA should create.
                If ``None``, all available cores will be used.
        """
        if num_processes == 1:
            self._implementation = Serial(
                initial_population=initial_population,
                fitness_calculator=fitness_calculator,
                mutator=mutator,
                crosser=crosser,
                generation_selector=generation_selector,
                mutation_selector=mutation_selector,
                crossover_selector=crossover_selector,
                fitness_normalizer=fitness_normalizer,
                key_maker=key_maker,
                logger=logger,
            )

        else:
            self._implementation = Parallel(
                initial_population=initial_population,
                fitness_calculator=fitness_calculator,
                mutator=mutator,
                crosser=crosser,
                generation_selector=generation_selector,
                mutation_selector=mutation_selector,
                crossover_selector=crossover_selector,
                fitness_normalizer=fitness_normalizer,
                key_maker=key_maker,
                logger=logger,
                num_processes=num_processes,
            )

    def get_generations(self, num_generations: int) -> Iterator[Generation[T]]:
        """
        Yield the generations of the evolutionary algorithm.

        Parameters:
            num_generations:
                The number of generations which should be yielded.
                Note that the initial population counts as a generation.
        Yields:
            A generation.
        """
        yield from self._implementation.get_generations(
            num_generations=num_generations,
        )



import numpy as np

from ...utilities import get_num_atom_ids, normalize_ids
from .utilities import get_maximum_diameter


def test_get_maximum_diameter(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_maximum_diameter`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the correct atomic
        positions of it atoms.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for
        :meth:`.Molecule.get_maximum_diameter`. This allows the
        testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    Notes
    -----
    This test compares the result of
    :meth:`.Molecule.get_maximum_diameter` to the result of
    :func:`.get_maximum_diameter`, which is a utility function defined
    for the purposes of this test. Because
    :func:`.get_maximum_diameter` is tested independently, in
    :mod:`.test_get_maximum_diameter_helper`, if its tests pass, then
    it can be assumed, that :func:`.get_maximum_diameter` gives
    correct results.

    Now, assuming that :func:`.get_maximum_diameter` passed all of its
    tests, this test compares the results of
    :meth:`.Molecule.get_maximum_diameter` to the results of
    :func:`.get_maximum_diameter`. If the results do not match, the
    fault can be placed on :meth:`.Molecule.get_maximum_diameter`,
    because :func:`.get_maximum_diameter` has already been verified to
    be correct by its own tests.

    """

    _test_get_maximum_diameter(
        molecule=case_data.molecule,
        get_atom_ids=get_atom_ids,
        maximum_diameter=get_maximum_diameter(
            position_matrix=case_data.position_matrix,
            atom_ids=tuple(
                normalize_ids(
                    molecule=case_data.molecule,
                    ids=get_atom_ids(case_data.molecule),
                )
            ),
        ),
    )


def _test_get_maximum_diameter(
    molecule,
    get_atom_ids,
    maximum_diameter,
):
    """
    Test :meth:`.Molecule.get_maximum_diameter`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for
        :meth:`.Molecule.get_maximum_diameter`. This allows the
        testing of different values of this parameter.

    maximum_diameter : :class:`float`
        The correct maximum_diameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    num_atom_ids = get_num_atom_ids(molecule, get_atom_ids)
    if num_atom_ids == 1:
        result = molecule.get_maximum_diameter(
            atom_ids=get_atom_ids(molecule),
        )
        assert result == 0
        return

    assert np.allclose(
        a=maximum_diameter,
        b=molecule.get_maximum_diameter(get_atom_ids(molecule)),
        atol=1e-32,
    )



from __future__ import annotations

import pymongo
import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def molecules() -> tuple[stk.ConstructedMolecule, ...]:
    bb1 = stk.BuildingBlock("BrC#CBr", [stk.BromoFactory()])
    bb2 = stk.BuildingBlock("BrCCBr", [stk.BromoFactory()])
    bb3 = stk.BuildingBlock("BrCCCBr", [stk.BromoFactory()])
    bb4 = stk.BuildingBlock("BrCCCCBr", [stk.BromoFactory()])
    bb5 = stk.BuildingBlock("BrCNCBr", [stk.BromoFactory()])
    bb6 = stk.BuildingBlock("BrCCNCBr", [stk.BromoFactory()])

    return (
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1,),
                repeating_unit="A",
                num_repeating_units=3,
            ),
        ),
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                num_repeating_units=3,
            ),
        ),
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb3,),
                repeating_unit="A",
                num_repeating_units=3,
            ),
        ),
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb3, bb4),
                repeating_unit="AB",
                num_repeating_units=3,
            ),
        ),
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb5,),
                repeating_unit="A",
                num_repeating_units=3,
            ),
        ),
        stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb5, bb6),
                repeating_unit="AB",
                num_repeating_units=3,
            ),
        ),
    )


def get_database(
    database_name: str,
    mongo_client: pymongo.MongoClient,
    key_makers: tuple[stk.MoleculeKeyMaker, ...],
    indices: tuple[str, ...],
) -> stk.ConstructedMoleculeMongoDb:
    return stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        jsonizer=stk.ConstructedMoleculeJsonizer(key_makers),
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        indices=indices,
    )


@pytest.fixture
def constructed_molecule_mongo_db(
    mongo_client: pymongo.MongoClient,
    molecules: tuple[stk.ConstructedMolecule, ...],
) -> CaseData:
    inchi = stk.Inchi()
    smiles = stk.Smiles()
    database_name = "_test_get_all_constructed_molecules"
    mongo_client.drop_database(database_name)

    inchi_molecules = molecules[:2]
    smiles_molecules = molecules[2:4]
    inchi_and_smiles_molecules = molecules[4:]

    inchi_database = get_database(
        database_name=database_name,
        mongo_client=mongo_client,
        key_makers=(inchi,),
        indices=(inchi.get_key_name(),),
    )
    smiles_database = get_database(
        database_name=database_name,
        mongo_client=mongo_client,
        key_makers=(smiles,),
        indices=(smiles.get_key_name(),),
    )

    inchi_and_smiles_database = get_database(
        database_name=database_name,
        mongo_client=mongo_client,
        key_makers=(inchi, smiles),
        indices=(),
    )

    for molecule in inchi_molecules:
        inchi_database.put(molecule)

    for molecule in smiles_molecules:
        smiles_database.put(molecule)

    for molecule in inchi_and_smiles_molecules:
        inchi_and_smiles_database.put(molecule)

    inchi_key_database = get_database(
        database_name=database_name,
        mongo_client=mongo_client,
        key_makers=(stk.InchiKey(),),
        indices=(),
    )

    expected_molecules = {
        smiles.get_key(molecule): molecule for molecule in molecules
    }

    return CaseData(
        database=inchi_key_database,
        expected_molecules=expected_molecules,
    )



from functools import partial

import numpy as np
import stk
from scipy.spatial.distance import euclidean


def get_fg_position(id, building_block):
    (functional_group,) = building_block.get_functional_groups(id)
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_centroid(building_block):
    return building_block.get_centroid()


def get_closest_point(points, point):
    return min(points, key=partial(euclidean, point))


def get_edges(vertex, angle):
    position1, position2 = get_points(vertex.get_position(), angle)
    vertex2 = stk.Vertex(1, position1)
    vertex3 = stk.Vertex(2, position2)
    yield stk.Edge(0, vertex, vertex2)
    yield stk.Edge(1, vertex, vertex3)


def get_points(position, angle):
    rotation_matrix = stk.rotation_matrix_arbitrary_axis(
        angle=angle,
        axis=np.array([0, 0, 1], dtype=np.float64),
    )
    points = (
        rotation_matrix
        @ np.array(
            [
                [-10, 0, 0],
                [10, 0, 0],
            ],
            dtype=np.float64,
        ).T
    )

    yield from (position + point for point in points.T)



class ValueDatabase:
    """
    Abstract base class for storing of molecular property values.

    Examples
    --------
    *Subclass Implementation*

    The source of any of the subclasses, listed in
    :mod:`molecule_value_database <.databases.value>`,
    can serve as good examples.

    *Iterating Through Entries in the Database*

    The :meth:`.MoleculeDatabase.get_all` and
    :meth:`.ConstructedMoleculeDatabase.get_all` methods  can be used
    to iterate through all of the database's entries, which can then
    be used to access values in the :class:`.ValueDatabase`

    .. testsetup:: iterating-through-entries-in-the-database

        import stk

        # Change the database used, so that when a developer
        # runs the doctests locally, their "stk" database is not
        # contaminated.
        _test_database = '_stk_doctest_database'

        _old_value_init = stk.ValueMongoDb
        stk.ValueMongoDb = lambda mongo_client, collection: (
            _old_value_init(
                mongo_client=mongo_client,
                database=_test_database,
                collection=collection,
            )
        )

        _old_molecule_init = stk.MoleculeMongoDb
        stk.MoleculeMongoDb = lambda mongo_client: _old_molecule_init(
            mongo_client=mongo_client,
            database=_test_database,
        )

        # Change the database MongoClient will connect to.

        import os
        import pymongo

        _mongo_client = pymongo.MongoClient
        _mongodb_uri = os.environ.get(
            'MONGODB_URI',
            'mongodb://localhost:27017/'
        )
        pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

    .. testcode:: iterating-through-entries-in-the-database

        import pymongo

        client = pymongo.MongoClient()
        molecule_db = stk.MoleculeMongoDb(client)

        value_db = stk.ValueMongoDb(
            mongo_client=client,
            collection='atom_counts',
        )

        molecule = stk.BuildingBlock('BrBr')
        molecule_db.put(molecule)
        value_db.put(molecule, molecule.get_num_atoms())


        for molecule in molecule_db.get_all():
            try:
                value = value_db.get(molecule)
                print(value)
            except KeyError:
                # In case molecule is not in value_db.
                pass

    .. testoutput:: iterating-through-entries-in-the-database

        2

    .. testcleanup:: iterating-through-entries-in-the-database

        stk.ValueMongoDb = _old_value_init
        stk.MoleculeMongoDb = _old_molecule_init
        pymongo.MongoClient().drop_database(_test_database)
        pymongo.MongoClient = _mongo_client

    """

    def put(self, molecule, value):
        """
        Put a value into the database.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule which is associated with the `value`.

        value : :class:`object`
            Some value associated with `molecule`.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()

    def get(self, molecule):
        """
        Get the stored value for `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule whose value is to be retrieved from the
            database.

        Returns
        -------
        :class:`object`
            The value associated with `molecule`.

        Raises
        ------
        :class:`KeyError`
            If `molecule` is not found in the database.

        """

        raise NotImplementedError()



"""
Single Atom
===========

"""

from .generic_functional_group import GenericFunctionalGroup


class SingleAtom(GenericFunctionalGroup):
    """
    Represents an abstract single atom functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom]``.

    """

    def __init__(self, atom):
        """
        Initialize a :class:`.SingleAtom` instance.

        Parameters
        ----------
        atom : :class:`.Atom`
            Any :class:`.Atom` will work.

        """

        self._atom = atom
        atoms = (atom,)
        super().__init__(atoms, atoms, ())

    def get_atom(self):
        """
        Get the atom that defines the functional group.

        Returns
        -------
        :class:`.Atom`
            The atom.

        """

        return self._atom

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def clone(self):
        clone = super().clone()
        clone._atom = self._atom
        return clone

    def __repr__(self):
        return f"{self.__class__.__name__}({self._atom})"



import stk

from ...utilities import (
    DatabaseEntry,
    DatabaseState,
    assert_database_state,
)
from .utilities import get_database_state


def test_update_1(mongo_client):
    """
    Test that existing entries are updated.

    """

    collection = "_test_update_1"
    database_name = "_test_update_1"
    mongo_client.drop_database(database_name)

    database = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
    )

    molecule = stk.BuildingBlock("CCC")

    database.put(molecule, 12)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    v=12,
                ): 1,
            }
        ),
    )

    database.put(molecule, 43)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    v=43,
                ): 1,
            }
        ),
    )


def test_update_2(mongo_client):
    """
    Test that existing entries are updated.

    In this test, you first create two separate entries, using
    different molecule keys. You then update both at the same time,
    with a database which uses both molecule keys.

    """

    collection = "_test_update_2"
    database_name = "_test_update_2"
    mongo_client.drop_database(database_name)

    database1 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(stk.InchiKey(),),
    )
    database2 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(stk.Smiles(),),
    )
    database3 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )

    molecule = stk.BuildingBlock("CCC")

    database1.put(molecule, 12)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    v=12,
                ): 1,
            }
        ),
    )

    # Should add another entry, as a different key maker is used.
    database2.put(molecule, 32)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    v=12,
                ): 1,
                DatabaseEntry(
                    SMILES=stk.Smiles().get_key(molecule),
                    v=32,
                ): 1,
            }
        ),
    )

    # Should update both entries as both key makers are used.
    database3.put(molecule, 56)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    SMILES=stk.Smiles().get_key(molecule),
                    v=56,
                ): 2,
            }
        ),
    )


def test_update_3(mongo_client):
    """
    Test that existing entries are updated.

    In this test, you first create one entry with two keys. Then
    update the entry with databases, each using 1 different key.
    No duplicate entries should be made in the database this way.

    """

    collection = "_test_update_3"
    database_name = "_test_update_3"
    mongo_client.drop_database(database_name)

    database1 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )
    database2 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(stk.InchiKey(),),
    )
    database3 = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection=collection,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        key_makers=(stk.Smiles(),),
    )

    molecule = stk.BuildingBlock("CCC")

    database1.put(molecule, 12)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    SMILES=stk.Smiles().get_key(molecule),
                    v=12,
                ): 1
            }
        ),
    )

    # Should update the entry.
    database2.put(molecule, 32)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    SMILES=stk.Smiles().get_key(molecule),
                    v=32,
                ): 1,
            }
        ),
    )

    # Should also update the entry.
    database3.put(molecule, 62)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(
                    InChIKey=stk.InchiKey().get_key(molecule),
                    SMILES=stk.Smiles().get_key(molecule),
                    v=62,
                ): 1,
            }
        ),
    )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks={
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=[C+]1",
                            functional_groups=[stk.BromoFactory()],
                        ): (4, 5, 6, 7, 8),
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ): (9,),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+][C+2]C2("
                                "Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ): (0, 1, 2),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]"
                                "C2(Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ): (3,),
                    },
                ),
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)[C+]5["
                "C+2][C+]6[C+2]C7([C+2][CH+][C+]57)C5=C([C+]=[C+]5)[C+"
                "]5[C+2][C+]([C+2]C7([C+2][CH+][C+]57)C5=C2[C+]=[C+]5)"
                "C2=C([C+]=[C+]2)[C+]2[C+2][C+](C5=C4[C+]=[C+]5)[C+]4["
                "CH+][C+2]C4([C+2]2)C2=C6[C+]=[C+]2)[C+]13"
            ),
            name=name,
        ),
    ),
)
def cage_four_plus_six(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



def test_get_num_bonds(case_data):
    """
    Test :meth:`.Molecule.get_num_bonds`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the correct number
        of bonds.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.molecule.get_num_bonds() == case_data.num_bonds



from dataclasses import dataclass
from typing import Callable

import pymongo
import pytest
import rdkit.Chem.AllChem as rdkit
import stk

from ..case_data import CaseData


@dataclass(frozen=True)
class CaseDataData:
    """
    Data used to create a :class:`.CaseData` instance.

    Attributes
    ----------
    get_database : :class:`callable`
        Creates the database to test. Takes a
        :class:`pymongo.MongoClient` as input and returns a
        :class:`.MoleculeMongoDb` instance.

    molecule : :class:`.Molecule`
        The molecule to put and get from the database being tested.

    key : :class:`object`
        The key used to retrieve the :attr:`.molecule` from the
        database.

    """

    get_database: Callable[[pymongo.MongoClient], stk.ValueMongoDb]
    molecule: stk.Molecule
    key: object


@pytest.fixture(
    params=(
        lambda: CaseDataData(
            get_database=lambda mongo_client: (
                stk.ConstructedMoleculeMongoDb(
                    mongo_client=mongo_client,
                    database="_stk_test_database_for_testing",
                    put_lru_cache_size=0,
                    get_lru_cache_size=0,
                )
            ),
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrCCBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=2,
                ),
            ),
            key={
                "InChIKey": rdkit.MolToInchiKey(
                    rdkit.MolFromSmiles(SMILES="BrCCCCBr")
                ),
            },
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: (
                stk.ConstructedMoleculeMongoDb(
                    mongo_client=mongo_client,
                    database="_stk_test_database_for_testing",
                    put_lru_cache_size=128,
                    get_lru_cache_size=128,
                )
            ),
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrCCBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=2,
                ),
            ),
            key={
                "InChIKey": rdkit.MolToInchiKey(
                    rdkit.MolFromSmiles(SMILES="BrCCCCBr")
                ),
            },
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: (
                stk.ConstructedMoleculeMongoDb(
                    mongo_client=mongo_client,
                    database="_stk_test_database_for_testing",
                    jsonizer=stk.ConstructedMoleculeJsonizer(
                        key_makers=(
                            stk.MoleculeKeyMaker(
                                key_name="SMILES",
                                get_key=lambda molecule: (
                                    rdkit.MolToSmiles(
                                        mol=molecule.to_rdkit_mol(),
                                    )
                                ),
                            ),
                        ),
                    ),
                    put_lru_cache_size=0,
                    get_lru_cache_size=0,
                )
            ),
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="Br[C+2][C+2]Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=2,
                ),
            ),
            key={"SMILES": "Br[C+2][C+2][C+2][C+2]Br"},
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: (
                stk.ConstructedMoleculeMongoDb(
                    mongo_client=mongo_client,
                    database="_stk_test_database_for_testing",
                    jsonizer=stk.ConstructedMoleculeJsonizer(
                        key_makers=(
                            stk.MoleculeKeyMaker(
                                key_name="SMILES",
                                get_key=lambda molecule: (
                                    rdkit.MolToSmiles(
                                        mol=molecule.to_rdkit_mol(),
                                    )
                                ),
                            ),
                        ),
                    ),
                    put_lru_cache_size=128,
                    get_lru_cache_size=128,
                )
            ),
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="Br[C+2][C+2]Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=2,
                ),
            ),
            key={"SMILES": "Br[C+2][C+2][C+2][C+2]Br"},
        ),
    ),
)
def constructed_molecule_mongo_db(
    request,
    mongo_client: pymongo.MongoClient,
) -> CaseData:
    data = request.param()
    return CaseData(
        database=data.get_database(mongo_client),
        molecule=data.molecule,
        key=data.key,
    )



def test_get_key(case_data):
    """
    Test :meth:`.get_key`.

    Parameters
    ----------
    case_data : :class:`.CaseDat`
        A test case. Holds the key maker to test and the correct key
        it should produce.

    Returns
    -------
    None : :class`NoneType`

    """

    _test_get_key(
        key_maker=case_data.key_maker,
        molecule=case_data.molecule,
        key=case_data.key,
    )


def _test_get_key(key_maker, molecule, key):
    """
    Test :meth:`.get_key`.

    Parameters
    ----------
    key_maker : :class:`.MoleculeKeyMaker`
        The key maker to test.

    molecule : :class:`.Molecule`
        The molecule to pass to the `key_maker`.

    key : :class:`object`
        The correct key of `molecule`.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert key_maker.get_key(molecule) == key



import numpy as np
import pytest
import stk

from ....case_data import CaseData
from .utilities import (
    get_closest_point,
    get_edges,
    get_fg_position,
    get_points,
)


@pytest.fixture
def flip(position, building_block_2, angle):
    point2, point1 = points = tuple(get_points(position, angle))

    def get_fg0_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(0, building_block),
        )

    def get_fg1_point(building_block):
        return get_closest_point(
            points=points,
            point=get_fg_position(1, building_block),
        )

    vertex = stk.macrocycle.CycleVertex(
        id=0,
        position=position,
        flip=True,
        angle=angle + np.pi / 2,
    )
    return CaseData(
        vertex=vertex,
        edges=tuple(get_edges(vertex, angle)),
        building_block=building_block_2,
        position=position,
        alignment_tests={
            get_fg0_point: point1,
            get_fg1_point: point2,
        },
        functional_group_edges={0: 1, 1: 0},
    )



import numpy as np
import pytest

from ..utilities import get_centroid


@pytest.mark.parametrize(
    argnames=("position_matrix", "atom_ids", "centroid"),
    argvalues=(
        (
            np.array(
                [
                    [1.0, 2.0, 3.0],
                    [0.0, 0.0, 0.0],
                    [-1.0, -2.0, -3.0],
                ]
            ),
            (0, 1, 2),
            np.array([0.0, 0.0, 0.0]),
        ),
        (
            np.array(
                [
                    [1.0, 2.0, 3.0],
                    [0.0, 0.0, 0.0],
                    [-1.0, -2.0, -3.0],
                ]
            ),
            (0, 1),
            np.array([0.5, 1.0, 1.5]),
        ),
    ),
)
def test_get_centroid_helper(position_matrix, atom_ids, centroid):
    """
    Test :func:`.get_centroid`.

    Parameters
    ----------
    position_matrix : :class:`numpy.ndarray`
        A position matrix.

    atom_ids : :class:`tuple` of :class:`int`
        The ids of selected atoms.

    centroid : :class:`numpy.ndarray`
        The correct centroid.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.allclose(
        a=get_centroid(position_matrix, atom_ids),
        b=centroid,
        atol=1e-32,
    )



import typing

from stk._internal.ea.molecule_record import MoleculeRecord

T = typing.TypeVar("T", bound=MoleculeRecord)


class FitnessCalculator(typing.Generic[T]):
    """
    Abstract base class for fitness value calculators.

    Examples:

        *Subclass Implementation*

        You only need to implement :meth:`.get_fitness_value`.
    """

    def get_fitness_value(self, record: T) -> typing.Any:
        """
        Return the fitness value of a molecule.

        Parameters:
            record:
                The molecule whose fitness value should be calculated.
        Returns:
            The fitness value.
        """
        raise NotImplementedError()



import typing
from typing import Any

T = typing.TypeVar("T")


class FitnessNormalizer(typing.Generic[T]):
    """
    Abstract base class for fitness normalizers.

    A fitness normalizer takes a dictionary mapping
    :class:`.MoleculeRecord` instances to their fitness values
    and returns a new dictionary mapping them to their normalized
    fitness values. The primary benefit of a normalizer vs a
    :class:`.FitnessCalculator` is that a :class:`.FitnessNormalizer`
    has access to all members in the population when it is calculating
    the normalized fitness value, whereas a :class:`.FitnessCalculator`
    does not.
    """

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        """
        Normalize some fitness values.

        Parameters:
            fitness_values:
                The molecules which need to have their fitness values
                normalized.
        Returns:
            The new fitness value for each molecule.
        """
        raise NotImplementedError()



"""
Two Plus Two
============

"""

import numpy as np

from stk._internal.topology_graphs.cage.cage import Cage
from stk._internal.topology_graphs.edge import Edge

from .vertices import NonLinearVertex


class TwoPlusTwo(Cage):
    """
    Represents a tetrahedron cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.TwoPlusTwo((bb, )),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.TwoPlusTwo(
                building_blocks=(bb, ),
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Building blocks with three functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 3

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        NonLinearVertex(0, [_x, 0, -_x / np.sqrt(2)], False),
        NonLinearVertex(1, [-_x, 0, -_x / np.sqrt(2)], False),
        NonLinearVertex(2, [0, _x, _x / np.sqrt(2)], False),
        NonLinearVertex(3, [0, -_x, _x / np.sqrt(2)], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(5, _vertex_prototypes[2], _vertex_prototypes[3]),
    )

    _num_windows = 4
    _num_window_types = 1



import numpy as np
import stk

from ..utilities import is_clone


def test_with_scale(case_data):
    """
    Test :meth:`.Edge.with_scale`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the starting position, the scale to apply
        and the target position.

    Returns
    -------
    None : :class:`NoneType`

    """

    edge = stk.Edge(
        id=0,
        vertex1=stk.Vertex(0, [0, 0, 0]),
        vertex2=stk.Vertex(1, [10, 0, 0]),
        position=case_data.start,
    )
    # Save a clone to check immutability.
    clone = edge.clone()
    _test_with_scale(edge, case_data.scale, case_data.target)
    is_clone(edge, clone)


def _test_with_scale(edge, scale, target):
    """
    Teste :meth:`.Edge.with_scale`.

    Parameters
    ----------
    edge : :class:`.Edge`
        The edge to test.

    scale : :class:`float` or :class:`numpy.ndarray`
        The scale to apply to `edge`.

    target : :class:`numpy.ndarray`
        The correct position of the new edge.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = edge.with_scale(scale)
    assert np.allclose(
        a=clone.get_position(),
        b=target,
        atol=1e-16,
    )



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule which should be canonically ordered.

    result : :class:`.Molecule`
        What the molecule should look like after canonical ordering.

    """

    def __init__(self, molecule, result):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule which should be canonically ordered.

        result : :class:`.Molecule`
            What the molecule should look like after canonical
            ordering.

        """

        self.molecule = molecule
        self.result = result



import pytest
import stk

from ..case_data import CaseData


def has_bromo(building_block):
    (fg,) = building_block.get_functional_groups(0)
    return fg.__class__ is stk.Bromo


def _get_case_data_1() -> CaseData:
    bb1 = stk.BuildingBlock("BrCCBr", [stk.BromoFactory()])
    graph1 = stk.polymer.Linear((bb1,), "A", 2)

    bb2 = stk.BuildingBlock("BrCNCBr", [stk.BromoFactory()])
    graph2 = stk.polymer.Linear((bb2,), "A", 2)

    return CaseData(
        mutator=stk.RandomBuildingBlock(
            building_blocks=(bb2,),
            is_replaceable=has_bromo,
        ),
        record=stk.MoleculeRecord(graph1),
        mutation_record=stk.MutationRecord(
            molecule_record=stk.MoleculeRecord(graph2),
            mutator_name="RandomBuildingBlock",
        ),
    )


@pytest.fixture(
    scope="session",
    params=(_get_case_data_1,),
)
def random_building_block(request) -> CaseData:
    return request.param()



"""
M4L4 Square
===========

"""

from stk._internal.optimizers.null import NullOptimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import AngledVertex, LinearVertex


class M4L4Square(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L4Square(
                corners=bb1,
                linkers=bb2,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(2)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC=NC=C2)C3=CC=NC=C3'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L4Square(
                corners=bb1,
                linkers=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Both `corner` and `linker` vertices require building blocks with
    two functional groups for this topology. This class replaces the
    `building_blocks` parameter with the `corner` and `linker`
    parameters.

    See :class:`.Cage` for more details and examples.

    """

    def __init__(
        self,
        corners,
        linkers,
        vertex_alignments=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
        optimizer=NullOptimizer(),
    ):
        """
        Initialize a :class:`.M4L4Square`.

        Parameters
        ----------
        corners : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all corner vertices on the topology
            graph.

        linkers : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all linker vertices on the topology
            graph.

        vertex_alignments : :class:`dict`, optional
            A mapping from the id of a :class:`.Vertex`
            to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is referred to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        reaction_factory : :class:`.ReactionFactory`, optional
            The reaction factory to use for creating bonds between
            building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        optimizer : :class:`.Optimizer`, optional
            Used to optimize the structure of the constructed
            molecule.

        """

        if isinstance(corners, dict):
            building_blocks = corners
        else:
            building_blocks = {corners: (0, 1, 2, 3)}

        if isinstance(linkers, dict):
            linkers_dict = linkers
        else:
            linkers_dict = {linkers: (4, 5, 6, 7)}

        building_blocks.update(
            (building_block, vertices)
            for building_block, vertices in linkers_dict.items()
        )

        super().__init__(
            building_blocks,
            vertex_alignments=vertex_alignments,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
            optimizer=optimizer,
        )

    _vertex_prototypes = (
        AngledVertex(0, [1, 1, 0]),
        AngledVertex(1, [1, -1, 0]),
        AngledVertex(2, [-1, -1, 0]),
        AngledVertex(3, [-1, 1, 0]),
        LinearVertex(4, [1, 0, 0], False),
        LinearVertex(5, [0, -1, 0], False),
        LinearVertex(6, [-1, 0, 0], False),
        LinearVertex(7, [0, 1, 0], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(2, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(3, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(4, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(5, _vertex_prototypes[3], _vertex_prototypes[6]),
        Edge(6, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(7, _vertex_prototypes[0], _vertex_prototypes[7]),
    )

    _num_windows = 1
    _num_window_types = 1



"""
PDB Updating Utilities
======================

"""

import rdkit.Chem.AllChem as rdkit

from stk._internal.utilities.utilities import remake


def _with_structure_from_pdb(self, path):
    """
    Change structure to match a ``.pdb`` file.

    Parameters
    ----------
    path : :class:`str`
        The full path of the ``.pdb`` file from which the structure
        should be updated.

    Returns
    -------
    :class:`.Molecule`
        The molecule.

    """

    molecule = remake(
        rdkit.MolFromPDBFile(
            molFileName=path,
            sanitize=False,
            removeHs=False,
        )
    )
    return self._with_position_matrix(
        position_matrix=molecule.GetConformer().GetPositions()
    )



import itertools as it


def test_get_placer_ids(case_data):
    """
    Test :meth:`.BuildingBlock.get_placer_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the building block to test and the
        correct placer ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_placer_ids(
        building_block=case_data.building_block,
        placer_ids=case_data.placer_ids,
    )


def _test_get_placer_ids(building_block, placer_ids):
    """
    Test :meth:`.BuildingBlock.get_placer_ids`.

    Parameters
    ----------
    building_block : :class:`.BuildingBlock`
        The building block to test.

    placer_ids : :class:`tuple` of :class:`int`
        The correct placer ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    for placer1, placer2 in it.zip_longest(
        building_block.get_placer_ids(),
        placer_ids,
    ):
        assert placer1 == placer2



"""
Macrocycle Vertices
===================

"""

from __future__ import annotations

import typing

import numpy as np
from scipy.spatial.distance import euclidean

from stk._internal.building_block import BuildingBlock
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.vertex import Vertex


class CycleVertex(Vertex):
    """
    Represents a vertex in a macrocycle.

    """

    def __init__(
        self,
        id: int,
        position: typing.Union[np.ndarray, tuple[float, float, float]],
        flip: bool,
        angle: float,
    ) -> None:
        """
        Initialize a :class:`.CycleVertex` instance.

        Parameters:

            id:
                The id of the vertex.

            position:
                The position of the vertex.

            flip:
                If ``True``, the orientation of building blocks placed
                by the vertex will be flipped.

            angle:
                The position of the vertex along the cycle, specified
                by the `angle`.

        """

        super().__init__(id, position)
        self._flip = flip
        self._angle = angle

    def clone(self) -> CycleVertex:
        clone = self._clone()
        clone._flip = self._flip
        clone._angle = self._angle
        return clone

    def get_flip(self) -> bool:
        """
        Return ``True`` if the vertex flips building blocks it places.

        Returns:

            ``True`` if the vertex flips building blocks it places.

        """

        return self._flip

    def place_building_block(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> np.ndarray:
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        fg0, fg1 = building_block.get_functional_groups()
        fg0_position = building_block.get_centroid(
            atom_ids=fg0.get_placer_ids(),
        )
        fg1_position = building_block.get_centroid(
            atom_ids=fg1.get_placer_ids(),
        )
        return (
            building_block.with_rotation_between_vectors(
                start=fg1_position - fg0_position,
                target=np.array([-1 if self._flip else 1, 0, 0]),
                origin=self._position,
            )
            .with_rotation_about_axis(
                angle=self._angle - (np.pi / 2),
                axis=np.array([0, 0, 1]),
                origin=self._position,
            )
            .get_position_matrix()
        )

    def map_functional_groups_to_edges(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> dict[int, int]:
        fg0_position = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )

        def fg0_distance(edge: Edge) -> float:
            return euclidean(edge.get_position(), fg0_position)

        edge0 = min(edges, key=fg0_distance)
        return {
            0: edge0.get_id(),
            1: (edges[1].get_id() if edge0 is edges[0] else edges[0].get_id()),
        }

    def __str__(self) -> str:
        return (
            f"Vertex(id={self._id}, "
            f"position={tuple(self._position.tolist())}, "
            f"flip={self._flip}, "
            f"angle={self._angle})"
        )



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def iodo(get_atom_ids):
    a, b = get_atom_ids(2)
    return _iodo(stk.I(a), stk.C(b))


def _iodo(iodine, atom):
    bonders = (atom,)
    deleters = (iodine,)
    return GenericCaseData(
        functional_group=stk.Iodo(
            iodine=iodine,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(iodine, atom),
        bonders=bonders,
        deleters=deleters,
    )



import numpy as np
import pytest
import stk

from ....case_data import CaseData


@pytest.fixture
def cof3(init_at_shifted_center, id, aligner_edge, cell, shift_params):
    return CaseData(
        vertex=init_at_shifted_center(
            id=id,
            vertices=shift_params.vertices,
            cell_shifts=shift_params.cell_shifts,
            lattice_constants=shift_params.lattice_constants,
            aligner_edge=aligner_edge,
            cell=cell,
        ),
        id=id,
        position=shift_params.position,
        cell=cell,
    )


@pytest.fixture(
    params=(
        stk.cof.LinearVertex.init_at_shifted_center,
        stk.cof.NonLinearVertex.init_at_shifted_center,
        stk.cof.UnaligningVertex.init_at_shifted_center,
    ),
)
def init_at_shifted_center(request):
    return request.param


class ShiftParams:
    def __init__(
        self,
        vertices,
        cell_shifts,
        lattice_constants,
        position,
    ):
        self.vertices = vertices
        self.cell_shifts = cell_shifts
        self.lattice_constants = lattice_constants
        self.position = position


@pytest.fixture(
    params=(
        lambda: ShiftParams(
            vertices=(
                stk.Vertex(1, [-1, 1, 2]),
                stk.Vertex(2, [1, -1, -2]),
            ),
            cell_shifts=(
                np.array([0, 0, 0]),
                np.array([0, 0, 0]),
            ),
            lattice_constants=(
                np.array([1, 0, 0]),
                np.array([0, 1, 0]),
                np.array([0, 0, 1]),
            ),
            position=np.array([0.0, 0.0, 0.0]),
        ),
        lambda: ShiftParams(
            vertices=(
                stk.Vertex(1, [-1, 1, 2]),
                stk.Vertex(2, [1, -1, 0.0]),
            ),
            cell_shifts=(
                np.array([10, -10, 10]),
                np.array([-10, 10, -10]),
            ),
            lattice_constants=(
                np.array([1, 0, 0]),
                np.array([0, 1, 0]),
                np.array([0, 0, 1]),
            ),
            position=np.array([0.0, 0.0, 1.0]),
        ),
    ),
)
def shift_params(request) -> ShiftParams:
    return request.param()



import typing

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch, BatchKey
from stk._internal.key_makers.molecule import MoleculeKeyMaker

T = typing.TypeVar("T", bound=MoleculeRecord)


class YieldedBatches(typing.Generic[T]):
    """
    Keeps track of batches yielded by :meth:`.Selector.select`.
    """

    __slots__ = ("_molecules", "_batches", "_num", "_key_maker")

    def __init__(self, key_maker: MoleculeKeyMaker) -> None:
        self._key_maker = key_maker

        # Has all molecules yielded by select().
        self._molecules: set[str] = set()
        # Has the identity_key() of all batches yielded by select().
        self._batches: set[BatchKey] = set()
        # Counts the total number of times select() has yielded.
        self._num = 0

    def update(self, batch: Batch[T]) -> typing.Self:
        """
        Update tracked data with a new `batch`.

        Parameters:
            batch:
                A batch yielded by :meth:`.Selector.select`.
        Returns:
            The data tracker.
        """
        molecules = (record.get_molecule() for record in batch)
        self._molecules.update(map(self._key_maker.get_key, molecules))
        self._batches.add(batch.get_identity_key())
        self._num += 1
        return self

    def get_num(self) -> int:
        """
        Get the number of times :meth:`.Selector.select` has yielded.

        Returns:
            The total number of times :meth:`.Selector.select` has
            yielded.
        """
        return self._num

    def is_yielded_batch(self, batch: Batch[T]) -> bool:
        """
        Check if `batch` has already been yielded.

        Parameters:
            batch:
                The batch to check.
        Returns:
            ``True`` if `batch` has already been yielded.
        """
        return batch.get_identity_key() in self._batches

    def is_unyielded_batch(self, batch: Batch[T]) -> bool:
        """
        Check if `batch` has not been yielded.

        Parameters:
            batch:
                The batch to check.
        Returns:
            ``True`` if `batch` has not been yielded.
        """
        return batch.get_identity_key() not in self._batches

    def has_yielded_molecules(self, batch: Batch[T]) -> bool:
        """
        Check if `batch` contains any previously yielded molecules.

        Parameters:
            batch:
                The batch to check.
        Returns:
            ``True`` if `batch` contains any molecules which have
            previously been yielded.
        """
        return any(
            self._key_maker.get_key(molecule) in self._molecules
            for molecule in (record.get_molecule() for record in batch)
        )

    def has_no_yielded_molecules(self, batch: Batch[T]) -> bool:
        """
        Check if `batch` consists only of unyielded molecules.

        Parameters:
            batch:
                The batch to check.
        Returns:
            ``True`` if `batch` does not have any previously yielded
            molecules.
        """
        return all(
            self._key_maker.get_key(molecule) not in self._molecules
            for molecule in (record.get_molecule() for record in batch)
        )



"""
M4L6 Tetrahedron
================

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import NonLinearVertex


class M4L6Tetrahedron(Cage):
    """
    Represents a cage topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        iron_oct_delta = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb1,
            ),
        )

        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=iron_oct_delta,
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L6Tetrahedron(
                building_blocks=(iron_oct_delta, ),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        iron_oct_delta = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb1,
                optimizer=stk.MCHammer(),
            ),
        )

        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=iron_oct_delta,
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L6Tetrahedron(
                building_blocks=(iron_oct_delta, ),
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    This topology directly connects the vertices of the tetrahedron.

    Building blocks with three functional groups are required for this
    topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 3

    Examples
    --------
    *Building a Metal-Organic Tetrahedron*

    Many metal-organic cages are built using a process called
    subcomponent self-assembly, which is a complex chemical process
    that occurs in solution. Here, we provide an example of an
    alchemical approach to building these types of cages. It is
    alchemical because the bonds formed during construction are not
    the same as the experimental reaction. Instead of forming bonds at
    the metal centre, we create bonds between disconnected ligands.
    Firstly, we define the disconnected linker molecule, where we have
    added a bromine atom at the disconnection to perform the cage
    reaction.

    .. testcode:: building-a-metal-organic-tetrahedron

        import stk

        # Define coordinating ligand with dummy bromine groups and
        # metal coordinating functional groups.
        bb1 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

    We then build the desired metal complex with this linker. This
    process completes all metal-ligand reactions performed during the
    subcomponent self-assembly process.

    .. testcode:: building-a-metal-organic-tetrahedron

        # Produce a Fe+2 atom with 6 functional groups.
        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        # Build iron complex with delta stereochemistry.
        iron_oct_delta = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb1,
            ),
        )

    We then redefine the metal complex building block based on its
    bromine functional groups, which becomes the metal-based
    building block of any cage formed by this process.

    .. testcode:: building-a-metal-organic-tetrahedron

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=iron_oct_delta,
            functional_groups=[stk.BromoFactory()],
        )

    Finally, we build the :class:`M4L6Tetrahedron` cage using this
    building block.

    .. testcode:: building-a-metal-organic-tetrahedron

        # Build an M4L6 Tetrahedron.
        cage2 = stk.ConstructedMolecule(
            topology_graph=stk.cage.M4L6Tetrahedron(
                building_blocks=(iron_oct_delta, ),
            ),
        )

    Importantly, in the case that the linker cannot be disconnected
    in a symmetrical fashion, we have provided the
    :class:`M4L6TetrahedronSpacer` topology, which has a spacer
    vertex between the metal vertices on the tetrahedron. See
    :class:`.M4L6TetrahedronSpacer` for an example of its usage.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1 / (2 * np.sqrt(2))
    _y = 0.5
    _vertex_prototypes = (
        NonLinearVertex(0, [_y, 0, -_x]),
        NonLinearVertex(1, [-_y, 0, -_x]),
        NonLinearVertex(2, [0, _y, _x]),
        NonLinearVertex(3, [0, -_y, _x]),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(5, _vertex_prototypes[2], _vertex_prototypes[3]),
    )

    _num_windows = 4
    _num_window_types = 1



import itertools
import logging
import typing
from collections.abc import Callable, Iterable, Iterator

from stk._internal.ea.crossover.molecule_crosser import MoleculeCrosser
from stk._internal.ea.crossover.record import CrossoverRecord
from stk._internal.ea.fitness_calculators.fitness_calculator import (
    FitnessCalculator,
)
from stk._internal.ea.fitness_normalizers.fitness_normalizer import (
    FitnessNormalizer,
)
from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.mutation.mutator import MoleculeMutator
from stk._internal.ea.mutation.record import MutationRecord
from stk._internal.ea.selection.batch import Batch
from stk._internal.ea.selection.selectors.selector import Selector
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from ...generation import FitnessValues, Generation

T = typing.TypeVar("T", bound=MoleculeRecord)
A = typing.TypeVar("A")
B = typing.TypeVar("B")
Map: typing.TypeAlias = Callable[[Callable[[A], B], Iterable[A]], B]


class Implementation(typing.Generic[T]):
    """
    An implementation of the default evolutionary algorithm.
    """

    def __init__(
        self,
        initial_population: Iterable[T],
        fitness_calculator: FitnessCalculator[T],
        mutator: MoleculeMutator[T],
        crosser: MoleculeCrosser[T],
        generation_selector: Selector[T],
        mutation_selector: Selector[T],
        crossover_selector: Selector[T],
        fitness_normalizer: FitnessNormalizer[T],
        key_maker: MoleculeKeyMaker,
        logger: logging.Logger,
    ) -> None:
        self._initial_population = tuple(initial_population)
        self._fitness_calculator = fitness_calculator
        self._mutator = mutator
        self._crosser = crosser
        self._generation_selector = generation_selector
        self._mutation_selector = mutation_selector
        self._crossover_selector = crossover_selector
        self._fitness_normalizer = fitness_normalizer
        self._key_maker = key_maker
        self._logger = logger

    def _get_generations(
        self,
        num_generations: int,
        map_: Map,
    ) -> Iterator[Generation[T]]:
        def get_mutation_record(batch: Batch[T]) -> MutationRecord[T] | None:
            return self._mutator.mutate(next(iter(batch)))

        def get_key(record: T) -> str:
            return self._key_maker.get_key(record.get_molecule())

        population, keys = dedupe(self._initial_population, get_key)

        self._logger.info("Calculating fitness values of initial population.")
        fitness_values = dict(
            zip(
                population,
                map_(self._fitness_calculator.get_fitness_value, population),
            )
        )
        normalized_fitness_values = self._fitness_normalizer.normalize(
            fitness_values=fitness_values,
        )
        yield Generation(
            fitness_values={
                record: FitnessValues(raw, normalized_fitness_values[record])
                for record, raw in fitness_values.items()
            },
            mutation_records=(),
            crossover_records=(),
        )

        for generation in range(1, num_generations):
            self._logger.info(f"Starting generation {generation}.")
            self._logger.info(f"Population size is {len(population)}.")

            self._logger.info("Doing crossovers.")
            crossover_records = tuple(
                self._get_crossover_records(normalized_fitness_values)
            )

            self._logger.info("Doing mutations.")
            mutation_records = tuple(
                record
                for record in map(
                    get_mutation_record,
                    self._mutation_selector.select(normalized_fitness_values),
                )
                if record is not None
            )

            self._logger.info("Calculating fitness values.")

            offspring, keys = dedupe(
                items=(
                    record.get_molecule_record()
                    for record in crossover_records
                ),
                get_key=get_key,
                seen=keys,
            )
            mutants, keys = dedupe(
                items=(
                    record.get_molecule_record() for record in mutation_records
                ),
                get_key=get_key,
                seen=keys,
            )
            fitness_values.update(
                zip(
                    itertools.chain(offspring, mutants),
                    map_(
                        self._fitness_calculator.get_fitness_value,
                        itertools.chain(offspring, mutants),
                    ),
                )
            )
            normalized_fitness_values = self._fitness_normalizer.normalize(
                fitness_values=fitness_values,
            )
            population, keys = dedupe(
                items=(
                    molecule_record
                    for molecule_record, in self._generation_selector.select(
                        population=normalized_fitness_values
                    )
                ),
                get_key=get_key,
            )
            fitness_values = {
                record: fitness_values[record] for record in population
            }
            normalized_fitness_values = {
                record: normalized_fitness_values[record]
                for record in population
            }
            yield Generation(
                fitness_values={
                    record: FitnessValues(
                        raw, normalized_fitness_values[record]
                    )
                    for record, raw in fitness_values.items()
                },
                mutation_records=mutation_records,
                crossover_records=crossover_records,
            )

    def _get_crossover_records(
        self,
        population: dict[T, float],
    ) -> Iterator[CrossoverRecord[T]]:
        for batch in self._crossover_selector.select(population):
            yield from self._crosser.cross(tuple(batch))


def dedupe(
    items: Iterable[A],
    get_key: Callable[[A], str],
    seen: set[str] | None = None,
) -> tuple[list[A], set[str]]:
    if seen is None:
        seen = set()
    unique = []
    for item in items:
        if (key := get_key(item)) not in seen:
            unique.append(item)
            seen.add(key)
    return unique, seen



class CaseData:
    """
    A test case.

    Attributes
    ----------
    jsonizer : :class:`.MoleculeJsonizer` or \
            :class:`.ConstructedMoleculeJsonizer`
        The jsonizer to test.

    molecule : :class:`.Molecule`
        The molecule to JSONize.

    json : :class:`dict`
        The JSON of :attr:`.molecule`.

    """

    def __init__(self, jsonizer, molecule, json):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        jsonizer : :class:`.MoleculeJsonizer` or \
                :class:`.ConstructedMoleculeJsonizer`
            The jsonizer to test.

        molecule : :class:`.Molecule`
            The molecule to JSONize.

        json : :class:`dict`
            The correct JSON of `molecule`.

        """

        self.jsonizer = jsonizer
        self.molecule = molecule
        self.json = json



import numpy as np
import stk

from ...utilities import (
    DatabaseEntry,
    DatabaseState,
    HashableDict,
    assert_database_state,
)
from .utilities import get_database_state


def to_hashable_matrix(json):
    json = dict(json)
    return {
        "m": tuple(tuple(row) for row in json.pop("m")),
        **json,
    }


def to_hashable_constructed_molecule(json):
    json = dict(json)
    return {
        "BB": tuple(map(HashableDict, json.pop("BB"))),
        "aI": tuple(tuple(v) for v in json.pop("aI")),
        "bI": tuple(tuple(v) for v in json.pop("bI")),
        "nBB": tuple(json.pop("nBB")),
        **json,
    }


def test_update_1(mongo_client):
    """
    Test that existing entries are updated.

    """

    database_name = "_test_update_1"
    mongo_client.drop_database(database_name)

    database = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
    )
    jsonizer = stk.ConstructedMoleculeJsonizer()

    molecule = stk.BuildingBlock(
        smiles="BrCCBr",
        functional_groups=[stk.BromoFactory()],
    ).with_canonical_atom_ordering()

    polymer = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            # Use it as a building block twice, to make sure it is
            # not repeatedly added to the molecules database.
            building_blocks=(molecule, molecule),
            repeating_unit="AB",
            num_repeating_units=2,
        ),
    ).with_canonical_atom_ordering()
    json = jsonizer.to_json(polymer)

    database.put(polymer)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                DatabaseEntry(**json["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json["matrix"])): 1,
                DatabaseEntry(**json["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )

    polymer2 = polymer.with_position_matrix(
        position_matrix=np.zeros((polymer.get_num_atoms(), 3)),
    )
    json2 = jsonizer.to_json(polymer2)

    database.put(polymer2)
    assert_database_state(
        state1=get_database_state(database),
        state2=DatabaseState(
            {
                DatabaseEntry(**json["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json2["matrix"])): 1,
                DatabaseEntry(**json["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )


def test_update_2(mongo_client):
    """
    Test that existing entries are updated.

    In this test, your first create two separate entries, using
    different molecule keys. You then update both at the same time,
    with a database which uses both molecule keys.

    """

    database_name = "_test_update_2"
    mongo_client.drop_database(database_name)

    jsonizer1 = stk.ConstructedMoleculeJsonizer(
        key_makers=(stk.InchiKey(),),
    )
    database1 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer1,
    )

    jsonizer2 = stk.ConstructedMoleculeJsonizer(
        key_makers=(stk.Smiles(),),
    )
    database2 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer2,
    )

    jsonizer3 = stk.ConstructedMoleculeJsonizer(
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )
    database3 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer3,
    )

    molecule = stk.BuildingBlock(
        smiles="BrCCCBr",
        functional_groups=[stk.BromoFactory()],
    ).with_canonical_atom_ordering()

    polymer1 = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            # Use it as a building block twice, to make sure it is
            # not repeatedly added to the molecules database.
            building_blocks=(molecule, molecule),
            repeating_unit="AB",
            num_repeating_units=2,
        ),
    ).with_canonical_atom_ordering()
    json1 = jsonizer1.to_json(polymer1)

    database1.put(polymer1)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json1["matrix"])): 1,
                DatabaseEntry(**json1["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json1["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json1["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )

    # Should add another entry, as a different key maker is used.
    polymer2 = polymer1.with_position_matrix(
        position_matrix=np.zeros((polymer1.get_num_atoms(), 3)),
    )
    json2 = jsonizer2.to_json(polymer2)

    database2.put(polymer2)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json1["matrix"])): 1,
                DatabaseEntry(**json1["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json1["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json1["constructedMolecule"],
                    )
                ): 1,
                DatabaseEntry(**json2["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json2["matrix"])): 1,
                DatabaseEntry(**json2["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json2["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json2["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )

    # Should update both entries.
    polymer3 = polymer1.with_position_matrix(
        position_matrix=np.zeros((polymer1.get_num_atoms(), 3)),
    )
    json3 = jsonizer3.to_json(polymer3)

    database3.put(polymer3)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json3["molecule"]): 2,
                DatabaseEntry(**to_hashable_matrix(json3["matrix"])): 2,
                DatabaseEntry(**json3["buildingBlocks"][0]["molecule"]): 2,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json3["buildingBlocks"][0]["matrix"],
                    )
                ): 2,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json3["constructedMolecule"],
                    )
                ): 2,
                DatabaseEntry(**json3["molecule"]): 2,
                DatabaseEntry(**to_hashable_matrix(json3["matrix"])): 2,
                DatabaseEntry(**json3["buildingBlocks"][0]["molecule"]): 2,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json3["buildingBlocks"][0]["matrix"],
                    )
                ): 2,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json3["constructedMolecule"],
                    )
                ): 2,
            }
        ),
    )


def test_update_3(mongo_client):
    """
    Test that existing entries are updated.

    In this test, your first create one entry with two keys. Then
    update the entry with databases, each using 1 different key.
    No duplicate entries should be made in the database this way.

    """

    database_name = "_test_update_3"
    mongo_client.drop_database(database_name)

    jsonizer1 = stk.ConstructedMoleculeJsonizer(
        key_makers=(
            stk.InchiKey(),
            stk.Smiles(),
        ),
    )
    database1 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer1,
    )

    jsonizer2 = stk.ConstructedMoleculeJsonizer(
        key_makers=(stk.InchiKey(),),
    )
    database2 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer2,
    )

    jsonizer3 = stk.ConstructedMoleculeJsonizer(
        key_makers=(stk.Smiles(),),
    )
    database3 = stk.ConstructedMoleculeMongoDb(
        mongo_client=mongo_client,
        database=database_name,
        put_lru_cache_size=0,
        get_lru_cache_size=0,
        jsonizer=jsonizer3,
    )

    molecule = stk.BuildingBlock(
        smiles="BrCCCBr",
        functional_groups=[stk.BromoFactory()],
    ).with_canonical_atom_ordering()

    polymer1 = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            # Use it as a building block twice, to make sure it is
            # not repeatedly added to the molecules database.
            building_blocks=(molecule, molecule),
            repeating_unit="AB",
            num_repeating_units=2,
        ),
    ).with_canonical_atom_ordering()
    json1 = jsonizer1.to_json(polymer1)

    database1.put(polymer1)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json1["matrix"])): 1,
                DatabaseEntry(**json1["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json1["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json1["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )

    # Should update the entry.
    polymer2 = polymer1.with_position_matrix(
        position_matrix=np.zeros((polymer1.get_num_atoms(), 3)),
    )
    json2 = jsonizer2.to_json(polymer2)
    json2["matrix"] = dict(json1["matrix"])
    json2["matrix"]["m"] = jsonizer2.to_json(polymer2)["matrix"]["m"]

    database2.put(polymer2)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json2["matrix"])): 1,
                DatabaseEntry(**json1["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json1["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json1["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )

    # Should also update the entry.
    polymer3 = polymer1.with_position_matrix(
        position_matrix=np.zeros((polymer1.get_num_atoms(), 3)),
    )
    json3 = jsonizer3.to_json(polymer3)
    json3["matrix"] = dict(json1["matrix"])
    json3["matrix"]["m"] = jsonizer3.to_json(polymer3)["matrix"]["m"]

    database3.put(polymer3)
    assert_database_state(
        state1=get_database_state(database1),
        state2=DatabaseState(
            {
                DatabaseEntry(**json1["molecule"]): 1,
                DatabaseEntry(**to_hashable_matrix(json3["matrix"])): 1,
                DatabaseEntry(**json1["buildingBlocks"][0]["molecule"]): 1,
                DatabaseEntry(
                    **to_hashable_matrix(
                        json=json1["buildingBlocks"][0]["matrix"],
                    )
                ): 1,
                DatabaseEntry(
                    **to_hashable_constructed_molecule(
                        json=json1["constructedMolecule"],
                    )
                ): 1,
            }
        ),
    )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture
def cof2(init_at_center, id, vertices_, aligner_edge, cell):
    return CaseData(
        vertex=init_at_center(id, vertices_, aligner_edge, cell),
        id=id,
        position=(sum(v.get_position() for v in vertices_) / len(vertices_)),
        cell=cell,
    )


@pytest.fixture(
    params=(
        stk.cof.LinearVertex.init_at_center,
        stk.cof.NonLinearVertex.init_at_center,
        stk.cof.UnaligningVertex.init_at_center,
    ),
)
def init_at_center(request):
    return request.param


@pytest.fixture(
    params=(
        lambda: (stk.Vertex(0, [1, 2, 3]),),
        lambda: (
            stk.Vertex(0, [1, 2, 3]),
            stk.Vertex(1, [-1, 2, -32]),
        ),
    ),
)
def vertices_(request) -> tuple[stk.Vertex, ...]:
    return request.param()



"""
One Plus One
============

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge
from stk._internal.utilities.utilities import get_acute_vector

from .cage import Cage
from .vertices import NonLinearVertex


class OnePlusOneVertex(NonLinearVertex):
    def __init__(
        self,
        id,
        position,
        edge_normal,
        use_neighbor_placement=True,
        aligner_edge=0,
    ):
        super().__init__(
            id=id,
            position=position,
            use_neighbor_placement=use_neighbor_placement,
            aligner_edge=aligner_edge,
        )
        self._edge_normal = np.array(edge_normal, dtype=np.float64)

    def clone(self):
        clone = super().clone()
        clone._edge_normal = np.array(self._edge_normal)
        return clone

    def place_building_block(self, building_block, edges):
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        placer_centroid = building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        )
        building_block = building_block.with_rotation_between_vectors(
            start=get_acute_vector(
                reference=core_centroid - placer_centroid,
                vector=building_block.get_plane_normal(
                    atom_ids=building_block.get_placer_ids(),
                ),
            ),
            target=self._edge_normal,
            origin=self._position,
        )
        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )
        start = fg_bonder_centroid - self._position
        edge_coord = edges[self._aligner_edge].get_position()
        building_block = building_block.with_rotation_to_minimize_angle(
            start=start,
            target=edge_coord - edge_centroid,
            axis=self._edge_normal,
            origin=self._position,
        )
        return building_block.get_position_matrix()


class OnePlusOne(Cage):
    """
    Represents a capsule cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='BrCC(CBr)CBr',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.OnePlusOne({bb1: 0, bb2: 1}),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='BrCC(CBr)CBr',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.OnePlusOne(
                building_blocks={bb1: 0, bb2: 1},
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Building blocks with three functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 1

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        OnePlusOneVertex(0, [_x, 0.0, 0.0], [1, 0, 0], False),
        OnePlusOneVertex(1, [-_x, 0.0, 0.0], [-1, 0, 0], False),
    )
    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            position=np.array([0.0, 1.0, 0.0]),
        ),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            position=np.array([0.0, -1.0, 1.0]),
        ),
        Edge(
            id=2,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            position=np.array([0.0, -1.0, -1.0]),
        ),
    )

    _num_windows = 3
    _num_window_types = 1



"""
Amide
=====

"""

from __future__ import annotations

import typing

from stk._internal.atom import Atom
from stk._internal.elements import C, H, N, O

from .generic_functional_group import GenericFunctionalGroup


class Amide(GenericFunctionalGroup):
    """
    Represents an amide functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][carbon](=[oxygen])[nitrogen]([hydrogen1])[hydrogen2]``.

    """

    def __init__(
        self,
        carbon: C,
        oxygen: O,  # noqa: Not an ambiguous name.
        nitrogen: N,
        hydrogen1: H,
        hydrogen2: H,
        atom: Atom,
        bonders: tuple[Atom, ...],
        deleters: tuple[Atom, ...],
        placers: typing.Optional[tuple[Atom, ...]] = None,
    ) -> None:
        """
        Initialize a :class:`.Amide` instance.

        Parameters:

            carbon:
                The ``[carbon]`` atom.

            oxygen:
                The ``[oxygen]`` atom.

            nitrogen:
                The ``[nitrogen]`` atom.

            hydrogen1:
                The ``[hydrogen1]`` atom.

            hydrogen2:
                The ``[hydrogen2]`` atom.

            atom:
                The ``[atom]`` atom.

            bonders:
                The bonder atoms.

            deleters:
                The deleter atoms.

            placers:
                The placer atoms. If ``None`` the `bonders` will be
                used.

        """

        self._carbon = carbon
        self._oxygen = oxygen
        self._nitrogen = nitrogen
        self._hydrogen1 = hydrogen1
        self._hydrogen2 = hydrogen2
        self._atom = atom
        atoms = (carbon, oxygen, nitrogen, hydrogen1, hydrogen2, atom)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_carbon(self) -> C:
        """
        Get the ``[carbon]`` atom.

        Returns:

            The ``[carbon]`` atom.

        """

        return self._carbon

    def get_oxygen(self) -> O:  # noqa: Not an ambiguous name.
        """
        Get the ``[oxygen]`` atom.

        Returns:

            The ``[oxygen]`` atom.

        """

        return self._oxygen

    def get_nitrogen(self) -> N:
        """
        Get the ``[nitrogen]`` atom.

        Returns:

            The ``[nitrogen]`` atom.

        """

        return self._nitrogen

    def get_hydrogen1(self) -> H:
        """
        Get the ``[hydrogen1]`` atom.

        Returns:

            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_hydrogen2(self) -> H:
        """
        Get the ``[hydrogen2]`` atom.

        Returns:

            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def get_atom(self) -> Atom:
        """
        Get the ``[atom]`` atom.

        Returns:

            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self) -> Amide:
        clone = super()._clone()
        clone._carbon = self._carbon
        clone._oxygen = self._oxygen
        clone._nitrogen = self._nitrogen
        clone._hydrogen1 = self._hydrogen1
        clone._hydrogen2 = self._hydrogen2
        clone._atom = self._atom
        return clone

    def _with_ids(
        self,
        id_map: dict[int, int],
    ) -> Amide:
        super()._with_ids(id_map)

        if (carbon_id := self._carbon.get_id()) in id_map:
            self._carbon = self._carbon.with_id(id_map[carbon_id])

        if (oxygen_id := self._oxygen.get_id()) in id_map:
            self._oxygen = self._oxygen.with_id(id_map[oxygen_id])

        if (nitrogen_id := self._nitrogen.get_id()) in id_map:
            self._nitrogen = self._nitrogen.with_id(
                id=id_map[nitrogen_id],
            )

        if (hydrogen1_id := self._hydrogen1.get_id()) in id_map:
            self._hydrogen1 = self._hydrogen1.with_id(
                id=id_map[hydrogen1_id],
            )

        if (hydrogen2_id := self._hydrogen2.get_id()) in id_map:
            self._hydrogen2 = self._hydrogen2.with_id(
                id=id_map[hydrogen2_id],
            )

        if (atom_id := self._atom.get_id()) in id_map:
            self._atom = self._atom.with_id(id_map[atom_id])

        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> Amide:
        return self.clone()._with_ids(id_map)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._carbon}, {self._oxygen}, {self._nitrogen}, "
            f"{self._hydrogen1}, {self._hydrogen2}, {self._atom}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



import itertools as it

from tests.utilities import is_equivalent_atom, is_equivalent_molecule


def are_equivalent_functional_groups(
    functional_groups1,
    functional_groups2,
):
    functional_groups = it.zip_longest(
        functional_groups1,
        functional_groups2,
    )
    for fg1, fg2 in functional_groups:
        is_equivalent_functional_group(fg1, fg2)


def is_equivalent_functional_group(
    functional_group1,
    functional_group2,
):
    assert functional_group1.__class__ is functional_group2.__class__

    atoms = it.zip_longest(
        functional_group1.get_atoms(),
        functional_group2.get_atoms(),
    )
    for atom1, atom2 in atoms:
        is_equivalent_atom(atom1, atom2)

    for placer_id1, placer_id2 in it.zip_longest(
        functional_group1.get_placer_ids(),
        functional_group2.get_placer_ids(),
    ):
        assert placer_id1 == placer_id2

    for core_atom_id1, core_atom_id2 in it.zip_longest(
        functional_group1.get_core_atom_ids(),
        functional_group2.get_core_atom_ids(),
    ):
        assert core_atom_id1 == core_atom_id2


def is_equivalent_building_block(building_block1, building_block2):
    is_equivalent_molecule(building_block1, building_block2)
    are_equivalent_functional_groups(
        functional_groups1=building_block1.get_functional_groups(),
        functional_groups2=building_block2.get_functional_groups(),
    )
    for placer_id1, placer_id2 in it.zip_longest(
        building_block1.get_placer_ids(),
        building_block2.get_placer_ids(),
    ):
        assert placer_id1 == placer_id2

    for core_atom_id1, core_atom_id2 in it.zip_longest(
        building_block1.get_core_atom_ids(),
        building_block2.get_core_atom_ids(),
    ):
        assert core_atom_id1 == core_atom_id2


def is_clone_building_block(building_block1, building_block2):
    assert building_block1 is not building_block2
    is_equivalent_building_block(building_block1, building_block2)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule
        The molecule to test.

    num_bonds : :class:`int`
        The correct number of bonds of :attr:`.molecule`.

    """

    def __init__(self, molecule, num_bonds):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        num_bonds : :class:`int`
            The correct number of bonds of `molecule`.

        """

        self.molecule = molecule
        self.num_bonds = num_bonds



from .m2l4_lantern import *  # noqa
from .m3l3_triangle import *  # noqa
from .m3l6 import *  # noqa
from .m4l4_square import *  # noqa
from .m4l4_tetrahedron import *  # noqa
from .m4l6_tetrahedron import *  # noqa
from .m4l6_tetrahedron_spacer import *  # noqa
from .m4l8 import *  # noqa
from .m6l2l3_prism import *  # noqa
from .m6l12_cube import *  # noqa
from .m8l6_cube import *  # noqa
from .m12l24 import *  # noqa
from .m24l48 import *  # noqa



from .utilities import are_equivalent_atoms


def test_get_deleted_atoms(case_data):
    """
    Test correction deletion of atoms by a :class:`.Reaction`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Includes the reaction and atoms which should have
        been deleted.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_deleted_atoms(
        reaction_result=case_data.reaction.get_result(),
        deleted_atoms=case_data.deleted_atoms,
    )


def _test_get_deleted_atoms(reaction_result, deleted_atoms):
    """
    Test that the correct atoms are deleted by `reaction_result`.

    Parameters
    ----------
    reaction_result : :class:`.ReactionResult`
        The result of a reaction.

    deleted_atoms : :class:`tuple` of :class:`.Atom`
        The atoms which should be deleted by the reaction.

    Returns
    -------
    None : :class:`NoneType`

    """

    are_equivalent_atoms(
        atoms1=reaction_result.get_deleted_atoms(),
        atoms2=deleted_atoms,
    )



"""
Hexagonal
=========

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cof import Cof
from .vertices import LinearVertex, NonLinearVertex


class Hexagonal(Cof):
    """
    Represents a hexagonal COF topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='Brc1c(Br)c(Br)c(Br)c(Br)c1Br',
            functional_groups=[stk.BromoFactory()],
        )

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.Hexagonal(
                building_blocks=(bb1, bb2),
                lattice_size=(2, 2, 1),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
            ),
        )

    ``Collapser(scale_steps=False)`` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='Brc1c(Br)c(Br)c(Br)c(Br)c1Br',
            functional_groups=[stk.BromoFactory()],
        )

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.Hexagonal(
                building_blocks=(bb1, bb2),
                lattice_size=(2, 2, 1),
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
            ),
        )

    Building blocks with six and two functional groups are required
    for this topology graph.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cof-topology-graph-examples`:
    *Multi-Building Block COF Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 6-functional groups: 0 to 3
        | 2-functional groups: 4 to 15

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1.0, 0.0, 0.0]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5 / 1.7321]),
    )

    _non_linears = (
        NonLinearVertex(0, (1 / 4) * _a + (1 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(1, (1 / 4) * _a + (3 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(2, (3 / 4) * _a + (1 / 4) * _b + (1 / 2) * _c),
        NonLinearVertex(3, (3 / 4) * _a + (3 / 4) * _b + (1 / 2) * _c),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=4,
            vertices=(_non_linears[0], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=5,
            vertices=(_non_linears[0], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=6,
            vertices=(_non_linears[1], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=7,
            vertices=(_non_linears[1], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=8,
            vertices=(_non_linears[2], _non_linears[3]),
        ),
        LinearVertex.init_at_shifted_center(
            id=9,
            vertices=(_non_linears[0], _non_linears[2]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=10,
            vertices=(_non_linears[0], _non_linears[1]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=11,
            vertices=(_non_linears[0], _non_linears[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=12,
            vertices=(_non_linears[2], _non_linears[1]),
            cell_shifts=((0, 0, 0), (1, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=13,
            vertices=(_non_linears[2], _non_linears[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=14,
            vertices=(_non_linears[1], _non_linears[3]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=15,
            vertices=(_non_linears[3], _non_linears[0]),
            cell_shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[2]),
        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[3]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(
            id=11,
            vertex1=_vertex_prototypes[9],
            vertex2=_vertex_prototypes[2],
            periodicity=(-1, 0, 0),
        ),
        Edge(12, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(
            id=13,
            vertex1=_vertex_prototypes[10],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),
        Edge(14, _vertex_prototypes[11], _vertex_prototypes[0]),
        Edge(
            id=15,
            vertex1=_vertex_prototypes[11],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),
        Edge(16, _vertex_prototypes[12], _vertex_prototypes[2]),
        Edge(
            id=17,
            vertex1=_vertex_prototypes[12],
            vertex2=_vertex_prototypes[1],
            periodicity=(1, -1, 0),
        ),
        Edge(18, _vertex_prototypes[13], _vertex_prototypes[2]),
        Edge(
            id=19,
            vertex1=_vertex_prototypes[13],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),
        Edge(20, _vertex_prototypes[14], _vertex_prototypes[1]),
        Edge(
            id=21,
            vertex1=_vertex_prototypes[14],
            vertex2=_vertex_prototypes[3],
            periodicity=(-1, 0, 0),
        ),
        Edge(22, _vertex_prototypes[15], _vertex_prototypes[3]),
        Edge(
            id=23,
            vertex1=_vertex_prototypes[15],
            vertex2=_vertex_prototypes[0],
            periodicity=(1, 0, 0),
        ),
    )



import numpy as np
import stk

from ...utilities import is_clone


def test_with_rotation_about_axis(molecule, angle, axis, get_origin):
    """
    Test :meth:`.Molecule.with_rotation_about_axis`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    angle : :class:`float`
        The angle of the rotation.

    axis : :class:`numpy.ndarray`
        The axis on which the rotation happens.

    get_origin : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `origin` parameter for this test. Using a :class:`callable`
        allows the origin for the test to be set to molecule-specific
        values, such as the centroid of the molecule being tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    origin = get_origin(molecule)
    original = rotational_space_positions(molecule, axis, origin)
    new = molecule.with_rotation_about_axis(angle, axis, origin)
    is_clone(new, molecule)
    rotated = rotational_space_positions(new, axis, origin)
    is_rotated(original, rotated, angle)


def is_rotated(original, rotated, angle):
    """
    Check that `rotated` is `original` with a rotation of `angle`.

    Parameters
    ----------
    original : :class:`numpy.ndarray`
        The position matrix of a molecule before rotation.

    rotated : :class:`numpy.ndarray`
        The position matrix of a molecule after rotation.

    angle : :class:`float`
        The rotational angle.

    Returns
    -------
    None : :class:`NoneType`

    """

    for atom_id in range(max(len(original), len(rotated))):
        # No rotation is expected if the atom was located at the
        # origin of the rotation.
        if np.allclose(original[atom_id], [0, 0, 0], 1e-13):
            assert np.allclose(
                a=original[atom_id],
                b=rotated[atom_id],
                atol=1e-13,
            )
        else:
            applied_rotation = stk.vector_angle(
                vector1=original[atom_id],
                vector2=rotated[atom_id],
            )
            assert abs(abs(angle) - applied_rotation) < 1e-13


def rotational_space_positions(molecule, axis, origin):
    """
    Get the atomic coordinates on the plane of the rotation.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule being rotated.

    axis : :class:`numpy.ndarray`
        The axis about which the rotation happens.

    origin : :class:`numpy.ndarray`
        The origin about which the rotation happens.

    Returns
    -------
    :class:`numpy.ndarray`
        An ``[n, 3]`` of atomic positions of `molecule`, projected
        onto the plane about which the rotation happens. The
        `axis` is the normal to this plane.

    """

    axis_matrix = np.repeat([axis], molecule.get_num_atoms(), 0).T
    positions = molecule.get_position_matrix() - origin
    return positions - (axis_matrix * (positions @ axis)).T



import stk


class CaseData:
    """
    A test case.

    Attributes
    ----------
    bond_info : :class:`.BondInfo`
        The bond info to test.

    bond : :class:`.Bond`
        The bond, which should be held by :attr:`.bond_info`.

    building_block : :class:`.BuildingBlock`
        The building block, which should be held by :attr:`.bond_info`.

    building_block_id : :class:`int`
        The building block id, which should be held by
        :attr:`.bond_info`.

    """

    def __init__(self, bond, building_block, building_block_id):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        bond : :class:`.Bond`
            The bond, which should be held by a bond info.

        building_block : :class:`.BuildingBlock`
            The building block, which should be held by a bond info.

        building_block_id : :class:`int`
            The building block id, which should be held by a bond info.

        """

        self.bond_info = stk.BondInfo(
            bond=bond,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self.bond = bond
        self.building_block = building_block
        self.building_block_id = building_block_id



import typing
from collections import abc
from functools import partial

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.construction_result.construction_result import (
    ConstructionResult,
)
from stk._internal.construction_state.construction_state import (
    ConstructionState,
)
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.edge_group import EdgeGroup
from stk._internal.topology_graphs.topology_graph.parallel import (
    _Parallel,
)
from stk._internal.topology_graphs.topology_graph.serial import (
    _Serial,
)
from stk._internal.topology_graphs.topology_graph.utilities import (
    _TopologyGraphImplementation,
)
from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import flatten


class TopologyGraph:
    """
    An abstract base class for topology graphs.

    It is responsible for the construction of molecules. To create
    a new topology graph, you want to subclass and implement this
    abstract base class.

    Notes:

        *Adding New Topology Graphs*

        You might notice that some of the methods of this abstract base
        class are implemented. This is purely for convenience when
        implementing subclasses. The implemented public methods are
        simply default implementations, which can safely be ignored or
        overridden, when implementing subclasses. Any private methods
        are implementation details of these default implementations.

        Many classes, such as :class:`.Vertex`, :class:`.Edge`,
        :class:`.EdgeGroup` and :class:`.ConstructionState`, exist as
        implementation details of this default :class:`.TopologyGraph`
        implementation. You could ignore all of them, and define a new
        :meth:`.construct` method from scratch. In fact, your topology
        graph does not have to be represented as a graph at all.
        However, using the default implementation of
        :class:`.TopologyGraph` makes it significantly easier to
        implement a construction process. When using the default
        implementation of :class:`.TopologyGraph`, you mostly just
        need to implement a :class:`.Vertex` subclass, which is much
        easier than figuring out the whole construction process from
        scratch. In addition, you get benefits like parallel
        construction for free, as it is included in the default
        implementation.

        Typically, adding a new topology graph will involve
        implementing any pure virtual methods of
        :class:`.TopologyGraph`, in a new subclass, as well as
        implementing any pure virtual methods of :class:`.Vertex`,
        again in a new subclass. Combined, this is just a handful of
        simple methods to implement. Sometimes, rarely, you might also
        want to subclass :class:`.ConstructionState`, when you want to
        add additional hooks during construction, by extending
        the methods of this class. If you do this, make sure to
        override :meth:`._get_construction_state` to return your
        subclass of :class:`.ConstructionState`, rather than the base
        class, as is done by default. You can subclass and extend the
        methods of any class as you wish, but it would be unusual if
        this doesn't cover all your requirements.

        *The Default Implementation*

        The default implementation of :class:`.TopologyGraph`
        represents the constructed molecule through a graph. The
        vertices indicate where building blocks are placed and the
        edges indicate which building blocks have bonds formed between
        them by the construction process.

        :class:`.Vertex` instances are responsible for placing the
        building block molecules. By initializing the vertices with
        different parameters, you can alter how they position the
        building block molecules, and therefore allow the user to
        easily specify a different structural isomer.

        Once a building block is placed on a vertex, the functional
        groups on the building block must be mapped to the different
        edges connected to the vertex. The number of functional groups
        in the building block must match the number of edges connected
        to the vertex.

        Once the functional groups are mapped to edges, the edges are
        used to perform reactions on the building blocks. Edges are
        grouped in an :class:`.EdgeGroup`, and all functional groups
        present in the edge group are reacted together. Normally,
        unless you are doing something very exotic, an
        :class:`.EdgeGroup` will hold just one :class:`.Edge`, and the
        two functional groups on that edge will be reacted together
        through a single :class:`.Reaction`. This reaction will
        normally add the bonds which are required to form the
        joined-up constructed molecule, but note that it does not have
        to add any bonds at all. In addition, a :class:`.Reaction` can
        add and remove atoms from the constructed molecule. Which
        reaction is selected to join the functional groups depends on
        the :class:`.ReactionFactory` given to the
        :class:`.TopologyGraph` during initialization.

        Once this is done, you have a :class:`.ConstructedMolecule`.

    Examples:

        *Subclass Implementation*

        The source code of subclasses, listed in
        :mod:`~.topology_graph.topology_graph.topology_graph`, can
        serve as good examples.

        *Changing the Building Blocks of a Topology Graph*

        To change the building blocks used by a topology graph you
        can use :meth:`.with_building_blocks` to get a clone of the
        topology graph holding the new building blocks

        .. testcode:: changing-the-building-blocks-of-a-topology-graph

            import stk

            bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
            bb2 = stk.BuildingBlock('BrCCCBr', [stk.BromoFactory()])

            linear = stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='A',
                num_repeating_units=15,
            )

            bb3 = stk.BuildingBlock('BrCNCBr', [stk.BromoFactory()])
            # All bb1 instances are replaced by bb3, but bb2 remains
            # in place.
            clone = linear.with_building_blocks({
                bb1: bb3,
            })
    """

    _implementation: _TopologyGraphImplementation

    def __init__(
        self,
        building_block_vertices: dict[
            BuildingBlock,
            abc.Sequence[Vertex],
        ],
        edges: tuple[Edge, ...],
        reaction_factory: ReactionFactory,
        construction_stages: tuple[
            abc.Callable[[Vertex], bool],
            ...,
        ],
        num_processes: int,
        optimizer: Optimizer,
        edge_groups: tuple[EdgeGroup, ...] | None = None,
    ) -> None:
        """
        Parameters:

            building_block_vertices:
                Maps each :class:`.BuildingBlock` to be placed, to the
                :class:`.Vertex` instances, on which it should be
                placed.

            edges:
                The edges which make up the topology graph.

            reaction_factory:
                Used to pick which :class:`.Reaction` is used on each
                :class:`.EdgeGroup` of the topology graph.

            construction_stages:
                A collection of :class:`~collections.abc.Callable`,
                each of which takes a :class:`.Vertex` and returns
                ``True`` or ``False``. If the first
                :class:`~collections.abc.Callable` is applied
                to a  vertex in the topology graph, and the result is
                ``True``, that vertex is a part of the first
                construction stage. The second :class:`callable` is
                then applied to all vertices not in the first stage
                and those which return ``True`` belong to the second
                stage and so on.

                Vertices which belong to the same construction stage
                all place building blocks together in parallel, before
                placement is done by any vertices which are part of a
                later stage. This breaks down parallel construction
                into serial stages if synchronization between stages
                is needed.

                If the topology graph is performing construction
                serially, then all vertices which belong to an earlier
                stage will place their building block before those at
                a later stage.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            edge_groups:
                The edge groups of the topology graph, if ``None``,
                every :class:`.Edge` is in its own edge group.
        """
        self._scale = scale = self._get_scale(building_block_vertices)

        def apply_scale(item):
            return item.with_scale(scale)

        self._building_block_vertices = {
            building_block: tuple(map(apply_scale, vertices))
            for building_block, vertices in building_block_vertices.items()
        }
        self._edges = tuple(map(apply_scale, edges))
        self._reaction_factory = reaction_factory

        if num_processes == 1:
            self._implementation = _Serial(
                stages=tuple(self._get_stages(construction_stages)),
            )
        else:
            self._implementation = _Parallel(
                stages=tuple(self._get_stages(construction_stages)),
                num_processes=num_processes,
            )

        if edge_groups is None:
            edge_groups = tuple(EdgeGroup((edge,)) for edge in self._edges)
        self._edge_groups = edge_groups

        self._optimizer = optimizer

    def _with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> typing.Self:
        """
        Modify the topology graph.

        """

        # The original scaling first needs to be removed, so that when
        # the scale is recalculated with the new building blocks, it
        # has the same starting geometry.
        def undo_scale(vertex):
            return vertex.with_scale(1 / self._scale)

        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]]
        building_block_vertices = {
            building_block_map.get(bb, bb): tuple(map(undo_scale, vertices))
            for bb, vertices in self._building_block_vertices.items()
        }
        scale = self._get_scale(building_block_vertices)

        def scale_vertex(vertex):
            return vertex.with_scale(scale)

        self._building_block_vertices = {
            building_block: tuple(map(scale_vertex, vertices))
            for building_block, vertices in building_block_vertices.items()
        }

        def scale_edge(edge):
            # Remove the old scale and apply the new one.
            return edge.with_scale(scale / self._scale)

        self._edges = edges = tuple(map(scale_edge, self._edges))

        def get_new_edge(edge_id):
            return edges[edge_id]

        self._edge_groups = tuple(
            EdgeGroup(map(get_new_edge, edge_group.get_edge_ids()))
            for edge_group in self._edge_groups
        )
        self._scale = scale
        return self

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> typing.Self:
        """
        Return a clone holding different building blocks.

        Parameters:

            building_block_map:
                Maps a building block in the current topology
                graph to the building block which should replace
                it in the clone. If a building block should be not
                replaced in the clone, it can be omitted from the map.

        Returns:
            The clone.
        """
        return self.clone()._with_building_blocks(building_block_map)

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            The clone.
        """
        return self._clone()

    def _clone(self) -> typing.Self:
        clone = self.__class__.__new__(self.__class__)
        clone._scale = self._scale
        clone._building_block_vertices = dict(self._building_block_vertices)
        clone._edges = self._edges
        clone._reaction_factory = self._reaction_factory
        clone._implementation = self._implementation
        clone._optimizer = self._optimizer
        clone._edge_groups = self._edge_groups
        return clone

    def get_building_blocks(self) -> abc.Iterator[BuildingBlock]:
        """
        Yield the building blocks.

        Building blocks are yielded in an order based on their
        position in the topology graph. For two equivalent
        topology graphs, but with different building blocks,
        equivalently positioned building blocks will be yielded at the
        same time.

        Yields:
            A building block of the topology graph.
        """
        vertex_building_blocks = {}
        num_vertices = 0
        for (
            building_block,
            vertices,
        ) in self._building_block_vertices.items():
            for vertex in vertices:
                num_vertices += 1
                vertex_building_blocks[vertex.get_id()] = building_block

        yielded = set()
        for vertex_id in range(num_vertices):
            building_block = vertex_building_blocks[vertex_id]
            if building_block not in yielded:
                yielded.add(building_block)
                yield building_block

    def get_num_building_block(
        self,
        building_block: BuildingBlock,
    ) -> int:
        """
        Get the number of times `building_block` is present.

        Parameters:
            building_block:
                The building block whose frequency in the topology
                graph is desired.

        Returns:
            The number of times `building_block` is present in the
            topology graph.
        """
        return len(self._building_block_vertices.get(building_block, []))

    def _get_lattice_constants(self) -> abc.Iterator[np.ndarray]:
        """
        Yield the lattice constants of the topology graph.

        The a, b and c lattice constants are yielded, in that order.

        By default, this is an empty generator.

        Yields:
            A lattice constant.
        """
        return
        yield

    def construct(self) -> ConstructionResult:
        """
        Construct a :class:`.ConstructedMolecule`.

        Returns:
            The data describing the :class:`.ConstructedMolecule`.
        """
        state = self._get_construction_state()
        state = self._place_building_blocks(state)
        state = self._run_reactions(state)
        state = self._optimizer.optimize(state)
        return self._get_construction_result(state)

    def _get_construction_result(
        self,
        state: ConstructionState,
    ) -> ConstructionResult:
        """
        Get the result of the construction.

        Parameters:
            state:
                The state of the molecule being constructed.

        Returns:
            The data describing the :class:`.ConstructedMolecule`.
        """
        return ConstructionResult(state)

    def _get_construction_state(self) -> ConstructionState:
        return ConstructionState(
            building_block_vertices=self._building_block_vertices,
            edges=self._edges,
            lattice_constants=tuple(
                np.array(constant, dtype=np.float64) * self._scale
                for constant in self._get_lattice_constants()
            ),
        )

    def _get_scale(
        self,
        building_block_vertices: dict[
            BuildingBlock,
            abc.Sequence[Vertex],
        ],
    ) -> float:
        """
        Get the scale, which should be applied to topology graph.

        The scale should be applied to the position of every vertex
        and edge of topology graph. This allows to graph to adjust
        based on the size of the building blocks.

        Parameters:
            building_block_vertices:
                Maps every :class:`.BuildingBlock` of the topology
                graph, to the :class:`.Vertex` instances it is meant
                to be placed on.

        Returns:
            The scale.
        """
        raise NotImplementedError()

    def _place_building_blocks(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        """
        Place the building blocks onto the vertices.

        Parameters:
            state:
                Holds data necessary to construct the molecule.

        Returns:
            The new construction state, updated to account for the
            placed building blocks.
        """
        return self._implementation._place_building_blocks(state)

    def _run_reactions(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        """
        Perform the reactions on the building blocks.

        Parameters:
            state:
                The current state of the construction process.

        Returns:
            The new construction state, updated to account for the
            reactions between building blocks.
        """
        get_reaction = partial(
            self._reaction_factory.get_reaction,
            state,
        )
        reactions = tuple(map(get_reaction, self._edge_groups))
        results = map(
            lambda reaction: reaction.get_result(),
            reactions,
        )
        return state.with_reaction_results(reactions, results)

    def _get_stages(
        self,
        construction_stages,
    ) -> abc.Iterator[tuple[int, ...]]:
        """
        Yield the parallelizable stages of construction.

        Yields:
            Vertices ids, which can be placed in parallel.
        """
        stages: tuple[list[int], ...] = tuple(
            [] for i in range(len(construction_stages) + 1)
        )
        vertices = flatten(self._building_block_vertices.values())
        for vertex in vertices:
            placed = False
            for i, stage in enumerate(construction_stages):
                if stage(vertex):
                    stages[i].append(vertex.get_id())
                    placed = True
                    break
            if not placed:
                stages[-1].append(vertex.get_id())
        yield from (tuple(stage) for stage in stages if stage)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        raise NotImplementedError()



"""
Porphyrin
=========

"""

from stk._internal.topology_graphs.edge import Edge

from .metal_complex import MetalComplex
from .vertices import MetalVertex, UnaligningVertex


class Porphyrin(MetalComplex):
    """
    Represents a metal complex topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )
        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=CC2=CC3=CC=C([N]3)C=C4C=CC'
                '(=N4)C=C5C=CC(=N5)C=C1[N]2'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )
        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.Porphyrin(
                metals=bb1,
                ligands=bb2,
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    complex.get_atoms(),
                    complex.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in complex.get_bonds()
            ),
        )

    Metal building blocks with at least four functional groups are
    required for this topology.

    Ligand building blocks with at least four functional group are
    required for this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, )
        | ligands: (0, )

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (MetalVertex(0, (0, 0, 0)),)
    _ligand_vertex_prototypes = (UnaligningVertex(1, (0, 0, 0)),)

    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(0.1, 0, 0),
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(0, 0.1, 0),
        ),
        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(-0.1, 0, 0),
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=(0, -0.1, 0),
        ),
    )
